from django.core.exceptions import ValidationError
from django.forms import forms
from django.forms.fields import MultipleChoiceField
from django.forms.widgets import CheckboxSelectMultiple
from django.utils.encoding import force_text
from django.utils.translation import ugettext as _

from .utils import (
    get_translations_versions_for_object, get_conflict_fks_versions,
)


class RecoverObjectWithTranslationForm(forms.Form):
    translations = MultipleChoiceField(
        required=True,
        widget=CheckboxSelectMultiple(),
        label=_('Translations to restore:'),
        help_text=_('Please select translations which would be restored.'))

    def __init__(self, *args, **kwargs):
        # prepare data for misc lookups
        self.revision = kwargs.pop('revision')
        self.obj = kwargs.pop('obj')
        self.version = kwargs.pop('version')
        self.resolve_conflicts = kwargs.pop('resolve_conflicts')
        self.placeholders = kwargs.pop('placeholders')
        # do not check object which needs to be recovered
        versions = self.revision.version_set.exclude(pk=self.version.pk)

        super(RecoverObjectWithTranslationForm, self).__init__(*args, **kwargs)

        translatable = hasattr(self.obj, 'translations')
        if translatable:
            translation_versions = get_translations_versions_for_object(
                self.obj, self.revision, versions)
            # update form
            choices = [(translation_version.pk, force_text(translation_version))
                       for translation_version in translation_versions]
            self.fields['translations'].choices = choices
        else:
            # do not show translations options if object is not translated
            self.fields.pop('translations')

    def clean(self):
        data = super(RecoverObjectWithTranslationForm, self).clean()
        # if there is self.resolve_conflicts do not count them as conflicts
        exclude = {
            'pk__in': [version.pk for version in
                       self.resolve_conflicts + self.placeholders]}
        conflict_fks_versions = get_conflict_fks_versions(
            self.obj, self.version, self.revision,
            exclude=exclude)
        if bool(conflict_fks_versions):
            raise ValidationError(
                _('Cannot restore object, there are conflicts!'),
                code='invalid')
        return data

    def save(self):
        # restore placeholders
        for placeholder_version in self.placeholders:
            # note that only placeholders are being reverted, assuming that
            # cms plugins that are related to this placeholder were not deleted
            placeholder_version.revert()

        # if there is self.resolve_conflicts revert those objects to avoid
        # integrity errors, because user cannot do that form admin
        # assume that that was prepared for us in admin view
        for conflict in self.resolve_conflicts:
            conflict.revert()

        # revert main object
        self.version.revert()

        # revert translations, if there is translations
        translations_pks = self.cleaned_data.get(
            'translations', []) if hasattr(self, 'cleaned_data') else []
        translation_versions = self.revision.version_set.filter(
            pk__in=translations_pks) if translations_pks else []
        for translation_version in translation_versions:
            translation_version.revert()



# -*- coding: utf-8 -*-

from __future__ import unicode_literals
from django.db import transaction

import reversion
from reversion.models import Version

from django.contrib import admin
from django.core.urlresolvers import reverse, NoReverseMatch

from cms import api

from aldryn_reversion.test_helpers.test_app.models import (
    SimpleRegistered, SimpleNoAdmin, SimpleFK, SimpleRequiredFK, BlankFK,
)

from .base import (
    HelperModelsObjectsSetupMixin, CMSRequestBasedMixin, ReversionBaseTestCase,
    get_version_for_object,
)

VERSION_INFO = "{content_type_name} #{object_id_int}: {object_repr}"
PLACEHOLDER_STR = RESOLVABLE_CONFLICT_STR = NON_RESOLVABLE_CONFLICT_STR = (
    VERSION_INFO)
VERSIONS_INFO = {
    'ph': PLACEHOLDER_STR,
    'conflict': RESOLVABLE_CONFLICT_STR,
    'non_resolavable': NON_RESOLVABLE_CONFLICT_STR
}
PLACEHOLDER_INFO = (
    'The following placeholders were deleted and will be restored')
CONFLICT_INFO = 'Please restore required related objects first'
NON_RESOLVABLE_CONFLICT_INFO = (
    'The following conflicts would be recovered automatically'
)
ALL_INFO_MESSAGES = [PLACEHOLDER_INFO, CONFLICT_INFO,
                     NON_RESOLVABLE_CONFLICT_INFO]
REVERT_BUTTON = RECOVER_BUTTON = (
    """<input type="submit" value="Yes, I'm sure" />""")


class AdminUtilsMixin(object):
    def setUp(self):
        super(AdminUtilsMixin, self).setUp()
        admin.autodiscover()
        self.admin_registry = admin.site._registry

    def get_admin_url_for_obj(self, obj, view_name, version=None):
        """
        Build admin view_name url for object.
        """
        opts = obj._meta
        url = reverse(
            'admin:{0}_{1}_{2}'.format(
                opts.app_label,
                obj._meta.model_name,
                view_name),
            args=[obj.pk if version is None else version.pk])
        return url

    def get_admin_instance_for_object(self, obj_or_version):
        """
        Return an admin instance for object or version.object_version.object
        if version was provided instead of object.
        """
        kls = (obj_or_version._meta.model
               if not type(obj_or_version) == Version
               else obj_or_version.content_type.model_class())

        admin_instance = self.admin_registry[kls]
        return admin_instance

    def build_string_args(self, version):
        """
        Prepare dict for building version info the same way it is specified in
        the template. Method is being used to populate prepared raw strings:
        PLACEHOLDER_STR, RESOLVABLE_CONFLICT_STR, NON_RESOLVABLE_CONFLICT_STR
        """
        return {'content_type_name': version.content_type.name.capitalize(),
                'object_id_int': version.object_id_int,
                'object_repr': version.object_repr}


class ReversionRecoverAdminTestCase(AdminUtilsMixin,
                                    CMSRequestBasedMixin,
                                    HelperModelsObjectsSetupMixin,
                                    ReversionBaseTestCase):

    def get_admin_url_for_obj(self, obj, view_name='recover', version=None):
        """
        Build admin view url for object, if view_name is not specified - builds
        url for recover view.
        """
        return super(ReversionRecoverAdminTestCase,
                     self).get_admin_url_for_obj(obj, view_name, version)

    def get_version(self, object_or_version):
        """
        Returns version or latest version for object, if object_or_version is
        object not reversion.models.Version.
        """
        return (object_or_version if type(object_or_version) == Version
                else get_version_for_object(object_or_version))

    def get_recover_view_response(self, obj_or_version, version_id=None,
                                  language='en'):
        """
        Returns result of admin recover_view processing (response) for given
        object (object_or_version).
        If object is a Version - uses version.object_version.object.
        If version_id was not provided - latest available version.id is used.
        """
        if version_id is None:
            version_id = str(self.get_version(obj_or_version).id)
        url = self.get_admin_url_for_obj(self.get_object(obj_or_version))

        request = self.get_su_request(language, url)
        admin_instance = self.get_admin_instance_for_object(obj_or_version)
        return admin_instance.recover_view(request, version_id)

    def post_recover_view_response(self, obj_or_version, language='en',
                                   version_id=None):
        """
        Mock a post request to admin view. Returns result of admin recover_view
        processing (response) for given object (object_or_version).
        If object is a Version - uses version.object_version.object.
        If version_id was not provided - latest available version.id is used.
        """
        if version_id is None:
            version_id = str(self.get_version(obj_or_version).id)
        url = self.get_admin_url_for_obj(self.get_object(obj_or_version))
        request = self.get_su_request(language, url)
        request.method = 'POST'
        admin_instance = self.get_admin_instance_for_object(obj_or_version)
        return admin_instance.recover_view(request, version_id)

    def check_info(self, response, info_string, versions=None):
        """
        Check that response contains info_string message,
        if versions are passed builds version info strings and checks that
        they were rendered. Expects versions to be a dict i.e
        { 'ph': placeholder_versions_list,
          'conflict': conflict_versions_list,
          'non_resolvable': non_resolvable_conflicts_versions_list
        }
        """
        self.assertContains(response, info_string)
        if versions:
            search_for = []
            for msg, versions_list in versions.items():
                version_info = VERSIONS_INFO[msg]
                search_for += [version_info.format(
                    **self.build_string_args(version)
                ) for version in versions_list]

            for info in search_for:
                self.assertContains(response, info)

    def test_recover_view_access(self):
        # test accessible for admin registered objects, should be greated than 4
        # characters, since '/en/' is 4, and actually not raises an exception
        self.assertGreater(
            len(self.get_admin_url_for_obj(self.simple_registered)),
            4)
        # test not accessible for not registered objects
        with self.assertRaises(NoReverseMatch):
            self.get_admin_url_for_obj(self.simple_no_admin)

    def test_recover_view_works_with_simple_objects(self):
        simple_version = get_version_for_object(self.simple_registered)
        self.simple_registered.delete()
        self.assertEqual(SimpleRegistered.objects.count(), 0)

        response = self.get_recover_view_response(simple_version)
        for info in ALL_INFO_MESSAGES:
            self.assertNotContains(response, info)

        # test recover button is accessible for simple object
        self.assertContains(response, RECOVER_BUTTON)
        # test recover is actually recovers object
        response = self.post_recover_view_response(simple_version)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(SimpleRegistered.objects.count(), 1)

    def test_recover_warnings_on_conflicts(self):
        # check that there is no warning on conflictable object
        # if there is no conflict
        simple_fk_version = get_version_for_object(self.simple_fk)
        self.simple_fk.delete()
        self.assertEqual(SimpleFK.objects.count(), 0)
        response = self.get_recover_view_response(
            simple_fk_version, version_id=str(simple_fk_version.id))
        for info in ALL_INFO_MESSAGES:
            self.assertNotContains(response, info)
        # check that in this case recover button is accessible
        self.assertContains(response, RECOVER_BUTTON)

        # check that recovered object is actually being restored
        response = self.post_recover_view_response(simple_fk_version)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(SimpleFK.objects.count(), 1)

        # make conflict which is solvable by user
        simple_registered_version = get_version_for_object(
            self.simple_registered)
        conflict_url = self.get_admin_url_for_obj(
            self.simple_registered, version=simple_registered_version)
        simple_required_fk_version = get_version_for_object(
            self.simple_required_fk)
        self.simple_registered.delete()
        self.simple_required_fk.delete()
        self.assertEqual(SimpleRegistered.objects.count(), 0)
        self.assertEqual(SimpleRequiredFK.objects.count(), 0)

        # check that there is a link to conflict
        response = self.get_recover_view_response(simple_required_fk_version)
        # check info messages.
        self.check_info(response, CONFLICT_INFO,
                        versions={'conflict': [simple_registered_version]})
        self.assertContains(response, conflict_url)
        # check that there is no recover button
        self.assertNotContains(response, RECOVER_BUTTON)

    # Test that if model has blank/null true conflicts are detected
    def test_recover_warnings_on_conflicts_for_blank_fk(self):
        # check that there is no warning on conflictable object
        # if there is no conflict
        blank_fk_version = get_version_for_object(self.blank_fk)
        self.blank_fk.delete()
        self.assertEqual(BlankFK.objects.count(), 0)
        response = self.get_recover_view_response(
            blank_fk_version, version_id=str(blank_fk_version.id))
        for info in ALL_INFO_MESSAGES:
            self.assertNotContains(response, info)
        # check that in this case recover button is accessible
        self.assertContains(response, RECOVER_BUTTON)

        # check that recovered object is actually being restored
        response = self.post_recover_view_response(blank_fk_version)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(BlankFK.objects.count(), 1)

        # since we have only one object we can do that =)
        self.blank_fk = BlankFK.objects.get()
        # make conflict which is solvable by user
        blank_fk_version = get_version_for_object(self.blank_fk)
        # since there might (or actually is) another revisions, use the same
        # revision as blank_fk_version, otherwise conflict url will mismatch
        simple_registered_version = get_version_for_object(
            self.blank_fk.simple_relation,
            revision_pk=blank_fk_version.revision.pk)
        conflict_url = self.get_admin_url_for_obj(
            self.blank_fk.simple_relation, version=simple_registered_version)
        # delete objects
        self.blank_fk.delete()
        self.assertEqual(BlankFK.objects.count(), 0)
        self.simple_registered.delete()
        self.assertEqual(SimpleRegistered.objects.count(), 0)

        # check that there is a link to conflict
        response = self.get_recover_view_response(blank_fk_version)
        # check info messages.
        self.check_info(response, CONFLICT_INFO,
                        versions={'conflict': [simple_registered_version]})
        self.assertContains(response, conflict_url)
        # check that there is no recover button
        self.assertNotContains(response, RECOVER_BUTTON)

    def test_recover_no_warnings_on_non_recoverable_by_user_conflict(self):
        # check that if there is no conflict - there is no warning

        simple_fk_version = get_version_for_object(self.simple_fk)
        self.simple_fk.delete()
        self.assertEqual(SimpleFK.objects.count(), 0)
        response = self.get_recover_view_response(simple_fk_version)
        for info in ALL_INFO_MESSAGES:
            self.assertNotContains(response, info)

        # check that recover button is accessible
        self.assertContains(response, RECOVER_BUTTON)

        # make conflict
        no_admin_verson = get_version_for_object(self.simple_no_admin)
        self.simple_no_admin.delete()
        # check that there is no conflict warning
        response = self.get_recover_view_response(simple_fk_version)
        self.assertNotContains(response, CONFLICT_INFO)
        self.assertContains(response, NON_RESOLVABLE_CONFLICT_INFO)
        # check that conflict object is listed among objects being recovered
        self.check_info(response, NON_RESOLVABLE_CONFLICT_INFO,
                        versions={'non_resolavable': [no_admin_verson]})
        # check that recover button is accessible
        self.assertContains(response, RECOVER_BUTTON)
        # check recovered objects are actually restored
        response = self.post_recover_view_response(simple_fk_version)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(SimpleFK.objects.count(), 1)
        self.assertEqual(SimpleNoAdmin.objects.count(), 1)


class ReversionRevisionAdminTestCase(AdminUtilsMixin,
                                     CMSRequestBasedMixin,
                                     HelperModelsObjectsSetupMixin,
                                     ReversionBaseTestCase):

    def get_admin_url_for_obj(self, obj, view_name, version=None):
        """
        Build admin view_name url for object.
        """
        opts = obj._meta
        url = reverse(
            'admin:{0}_{1}_{2}'.format(
                opts.app_label,
                obj._meta.model_name,
                view_name),
            args=[obj.pk, version.pk])
        return url

    def get_revision_view_response(self, obj, version, language='en'):
        url = self.get_admin_url_for_obj(obj, 'revision', version)
        request = self.get_su_request(language, url)
        admin_instance = self.get_admin_instance_for_object(obj)
        return admin_instance.revision_view(request,
                                            str(obj.pk),
                                            str(version.pk))

    def post_revision_veiw_response(self, obj, version, language='en'):
        url = self.get_admin_url_for_obj(obj, 'revision', version)
        request = self.get_su_request(language, url)
        request.method = 'POST'
        admin_instance = self.get_admin_instance_for_object(obj)
        return admin_instance.revision_view(request,
                                            str(obj.pk),
                                            str(version.pk))

    def test_revision_view_is_accessible(self):
        simple_registered_version = get_version_for_object(
            self.simple_registered)
        response = self.get_revision_view_response(
            self.simple_registered, simple_registered_version)
        for version in simple_registered_version.revision.version_set.all():
            self.assertContains(response, VERSION_INFO.format(
                **self.build_string_args(version)))
        self.assertContains(response, REVERT_BUTTON)

    def test_revision_view_reverts_object_to_selected_state(self):
        initial_position = self.simple_registered.position
        new_position = 99
        self.create_revision(self.simple_registered, position=new_position)
        self.assertNotEqual(initial_position, self.simple_registered.position)
        prev_version = reversion.get_for_object(self.simple_registered)[1]
        response = self.post_revision_veiw_response(
            self.simple_registered, prev_version)
        self.assertEqual(response.status_code, 302)
        self.simple_registered = SimpleRegistered.objects.get(
            pk=self.simple_registered.pk)
        self.assertEquals(self.simple_registered.position, initial_position)


class AdminUtilsMethodsTestCase(AdminUtilsMixin,
                                CMSRequestBasedMixin,
                                HelperModelsObjectsSetupMixin,
                                ReversionBaseTestCase):

    def test_create_aldryn_revision(self):
        # would be used to get admin instance
        with_placeholder_version = get_version_for_object(
            self.with_placeholder)

        admin_instance = self.get_admin_instance_for_object(
            with_placeholder_version)
        plugin = api.add_plugin(self.with_placeholder.content,
                                'TextPlugin', language='en')
        plugin.body = 'Initial text'
        plugin.save()
        # ensure there was no versions for plugin before
        self. assertEqual(reversion.get_for_object(plugin).count(), 0)
        with transaction.atomic():
            with reversion.create_revision():
                admin_instance._create_aldryn_revision(
                    plugin,
                    comment='New aldryn revision with initial plugin')
        # ensure there is at least one version after create aldryn revision
        self. assertEqual(reversion.get_for_object(plugin).count(), 1)
        new_plugin_text = 'test plugin content was changed'
        plugin.body = new_plugin_text
        plugin.save()
        with transaction.atomic():
            with reversion.create_revision():
                admin_instance._create_aldryn_revision(
                    plugin,
                    comment='New aldryn revision with initial plugin')

        # ensure there is at least one version after create aldryn revision
        self. assertEqual(reversion.get_for_object(plugin).count(), 2)
        latest_plugin = plugin._meta.model.objects.get(pk=plugin.pk)

        # ensure text is latest
        self.assertEqual(latest_plugin.body, new_plugin_text)
        # ensure text is initial if reverted to previous revision
        prev_version = reversion.get_for_object(self.with_placeholder)[1]
        prev_version.revision.revert()
        # refresh from db
        latest_plugin = plugin._meta.model.objects.get(pk=plugin.pk)
        # ensure plugin text was chagned. Note however that there might be
        # different paths to ensure that text is chagned for CMSPlugin
        # This only checks that plugin content (which is text plugin not the cms
        # is reverted, so be careful.
        self.assertEqual(latest_plugin.body, 'Initial text')



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import reversion
from reversion.models import Version
from cms.models import Placeholder

from ..utils import (
    object_is_translation, object_has_placeholders,
    get_placeholder_fields_names, get_fk_models,
    get_translations_versions_for_object, get_deleted_objects_versions,
    get_conflict_fks_versions, get_deleted_placeholders,
    get_deleted_placeholders_for_object,
    RecursiveRevisionConflictResolver,
)

from .base import (
    ReversionBaseTestCase, HelperModelsObjectsSetupMixin,
    get_version_for_object
)
from aldryn_reversion.test_helpers.test_app.models import (
    SimpleNoAdmin, SimpleRegistered, WithPlaceholder,
    SimpleFK, BlankFK, ComplexOneFK, MultiLevelFK,
)


class UtilsTestCase(HelperModelsObjectsSetupMixin, ReversionBaseTestCase):

    def test_object_is_translation(self):
        # false for original object
        self.assertFalse(object_is_translation(self.with_translation))
        # true for translation object
        translation = self.with_translation.translations.all()[0]
        self.assertTrue(object_is_translation(translation))

    def test_object_has_placeholders(self):
        self.assertFalse(object_has_placeholders(self.with_translation))
        self.assertTrue(object_has_placeholders(self.with_placeholder))

    def test_get_placeholder_fields_names(self):
        placeholder_name = 'content'
        result = get_placeholder_fields_names(self.with_placeholder)
        self.assertEqual(len(result), 1)
        self.assertIn(placeholder_name, result)

        # check that calling function for second time doesn't changes result
        result = get_placeholder_fields_names(self.with_placeholder)
        self.assertEqual(len(result), 1)

    def test_get_fk_models(self):
        # all fk's, blankness in models shouldn't affect the result
        result_all = get_fk_models(self.simple_fk)
        result_models = [relation['model'] for relation in result_all]
        self.assertEqual(len(result_all), 1)
        self.assertIn(SimpleNoAdmin, result_models)

        result_all = get_fk_models(self.blank_fk)
        result_models = [relation['model'] for relation in result_all]
        self.assertEqual(len(result_all), 1)
        self.assertIn(SimpleRegistered, result_models)

        # required fk's
        result_required = get_fk_models(self.simple_fk, blank=False)
        result_required_models_only = [relation['model']
                                       for relation in result_required]
        self.assertEqual(len(result_required), 1)
        self.assertIn(SimpleNoAdmin, result_required_models_only)

        result_required = get_fk_models(self.blank_fk, blank=False)
        self.assertEqual(len(result_required), 0)

        # not required fk's
        result_not_required = get_fk_models(self.blank_fk, blank=True)
        result_not_required_models = [relation['model']
                                      for relation in result_not_required]
        self.assertEqual(len(result_not_required), 1)
        self.assertIn(SimpleRegistered, result_not_required_models)

        result_not_required = get_fk_models(self.simple_fk, blank=True)
        self.assertEqual(len(result_not_required), 0)

    def test_get_translations_versions_for_object(self):
        # check that object has 2 translation already
        self.assertEqual(self.with_translation.translations.count(), 2)
        # get_for_object returns latest version first.
        version_2 = reversion.get_for_object(self.with_translation)[0]
        version_1 = reversion.get_for_object(self.with_translation)[1]
        revision_2 = version_2.revision
        revision_1 = version_1.revision
        # compare count of translations from latest revision to actual count
        result_revision_2 = get_translations_versions_for_object(
            self.with_translation, revision_2)
        result_revision_1 = get_translations_versions_for_object(
            self.with_translation, revision_1)

        self.assertEqual(len(result_revision_2), 2)
        self.assertEqual(len(result_revision_1), 1)

        # test with providing versions explicitly, should respect versions
        # over revision
        exclude_en_pks = reversion.get_for_object(
            self.with_translation.translations.filter(
                language_code='en').get()).values_list('pk', flat=True)
        versions = revision_2.version_set.all().exclude(
            pk__in=exclude_en_pks)
        # check that we have versions (should be object and german translation)
        self.assertEqual(len(versions), 2)
        result_versions = get_translations_versions_for_object(
            self.with_translation, revision_1, versions=versions)
        self.assertEqual(len(result_versions), 1)

    def test_get_deleted_objects_versions(self):
        blank_fk_version = reversion.get_for_object(self.blank_fk)[0]
        blank_fk_pk = self.blank_fk.pk
        # ensure that returns nothing for not deleted objects
        result = get_deleted_objects_versions(
            blank_fk_version.revision.version_set.all())
        self.assertEqual(len(result), 0)

        # delete and ensure that object was deleted.
        self.blank_fk.delete()
        self.assertEqual(
            BlankFK.objects.filter(pk=blank_fk_pk).count(), 0)
        # test against deleted object
        result = get_deleted_objects_versions(
            blank_fk_version.revision.version_set.all())
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].object_id_int, blank_fk_pk)

        # test with mixing 2 deleted object Versions
        simple_fk_version = reversion.get_for_object(self.simple_fk)[0]
        simple_fk_pk = self.simple_fk.pk
        self.simple_fk.delete()
        self.assertEqual(SimpleFK.objects.filter(pk=simple_fk_pk).count(), 0)
        # construct versions with this two objects and test
        blank_fk_versions_pks = [version.pk for version in
                                 blank_fk_version.revision.version_set.all()]
        simple_fk_versions_pks = [version.pk for version in
                                  simple_fk_version.revision.version_set.all()]
        both_obj_versions_pks = blank_fk_versions_pks + simple_fk_versions_pks
        custom_versions = Version.objects.filter(pk__in=both_obj_versions_pks)
        result = get_deleted_objects_versions(custom_versions)
        self.assertEqual(len(result), 2)

    def test_get_conflict_fks_versions_with_simple_models(self):
        # test with object that has no relations
        simple_no_admin_version = reversion.get_for_object(
            self.simple_registered)[0]
        result = get_conflict_fks_versions(
            self.simple_registered, simple_no_admin_version,
            simple_no_admin_version.revision)
        self.assertEqual(len(result), 0)

        # test with object that has no conflicts
        simple_fk_version = reversion.get_for_object(self.simple_fk)[0]
        result = get_conflict_fks_versions(
            self.simple_fk, simple_fk_version, simple_fk_version.revision)
        self.assertEqual(len(result), 0)

        # test with object that has conflicts
        self.simple_fk.delete()
        self.simple_no_admin.delete()
        result = get_conflict_fks_versions(
            self.simple_fk, simple_fk_version, simple_fk_version.revision)
        self.assertEqual(len(result), 1)

        # check that exclude excludes by excluding previous result version
        result = get_conflict_fks_versions(
            self.simple_fk, simple_fk_version, simple_fk_version.revision,
            exclude={'pk': result[0].pk})
        self.assertEqual(len(result), 0)

    def test_get_conflict_fks_versions_with_blank_fk_model(self):
        # test with no conflict
        bank_fk_version = reversion.get_for_object(self.blank_fk)[0]
        result = get_conflict_fks_versions(
            self.blank_fk, bank_fk_version, bank_fk_version.revision)
        self.assertEqual(len(result), 0)

        # test with conflict,
        self.blank_fk.delete()
        self.simple_registered.delete()
        result = get_conflict_fks_versions(
            self.blank_fk, bank_fk_version, bank_fk_version.revision)
        self.assertEqual(len(result), 1)

    def test_get_conflict_fks_versions_with_blank_fk_no_fk(self):
        # test with new blank fk that has no relations, should not have
        # conflicts after delete.
        new_blank_fk = self.create_with_revision(BlankFK)
        new_blank_fk_version = reversion.get_for_object(new_blank_fk)[0]
        result = get_conflict_fks_versions(
            new_blank_fk, new_blank_fk_version, new_blank_fk_version.revision)
        self.assertEqual(len(result), 0)

    def test_get_deleted_placeholders(self):
        with_placeholder_version = reversion.get_for_object(
            self.with_placeholder)[0]
        # test returns nothing if there are no deleted placeholders
        result = get_deleted_placeholders(with_placeholder_version.revision)
        self.assertEqual(len(result), 0)

        # delete placeholder
        placeholder_pk = self.with_placeholder.content.pk
        self.with_placeholder.content.delete()
        # check that placeholder was actually deleted.
        self.assertEqual(self.with_placeholder._meta.model.objects.filter(
            pk=placeholder_pk).count(), 0)

        result = get_deleted_placeholders(with_placeholder_version.revision)
        self.assertEqual(len(result), 1)

        # test with revision that has more then one deleted placeholder
        # (misc objects)
        complex_one_fk_version = reversion.get_for_object(
            self.complex_one_fk)[0]
        other_placeholder_pk = self.complex_one_fk.complex_content.pk
        self.complex_one_fk.complex_content.delete()
        self.assertEqual(self.with_placeholder._meta.model.objects.filter(
            pk__in=[placeholder_pk, other_placeholder_pk]).count(), 0)
        result = get_deleted_placeholders(complex_one_fk_version.revision)
        self.assertEqual(len(result), 2)

    def test_get_deleted_placeholders_for_object(self):
        # test that it not fails with object that has no placeholders
        simple_no_admin_version = reversion.get_for_object(
            self.simple_no_admin)[0]
        result = get_deleted_placeholders_for_object(
            self.with_placeholder, simple_no_admin_version.revision)
        self.assertEqual(len(result), 0)

        # test if no placeholders were deleted
        with_placeholder_version = reversion.get_for_object(
            self.with_placeholder)[0]
        result = get_deleted_placeholders_for_object(
            self.with_placeholder, with_placeholder_version.revision)
        self.assertEqual(len(result), 0)

        # test if placeholder was deleted for this object
        # delete placeholder
        placeholder_pk = self.with_placeholder.content.pk
        self.with_placeholder.content.delete()
        # check that placeholder was actually deleted.
        self.assertEqual(self.with_placeholder._meta.model.objects.filter(
            pk=placeholder_pk).count(), 0)

        result = get_deleted_placeholders_for_object(
            self.with_placeholder, with_placeholder_version.revision)
        self.assertEqual(len(result), 1)

        # test if placeholder was deleted for not related object (no result)
        complex_one_fk_version = reversion.get_for_object(
            self.complex_one_fk)[0]
        other_placeholder_pk = self.complex_one_fk.complex_content.pk
        self.complex_one_fk.complex_content.delete()
        self.assertEqual(self.with_placeholder._meta.model.objects.filter(
            pk__in=[placeholder_pk, other_placeholder_pk]).count(), 0)

        result = get_deleted_placeholders_for_object(
            self.with_placeholder, with_placeholder_version.revision)
        self.assertEqual(len(result), 1)

        result = get_deleted_placeholders_for_object(
            self.complex_one_fk, complex_one_fk_version.revision)
        self.assertEqual(len(result), 1)

    def test_recursive_resolver_resolve_conflicts(self):
        # test with no conflicts
        #  * no translations, placeholders, fks
        simple_no_adm_version = get_version_for_object(
            self.simple_registered)
        result = RecursiveRevisionConflictResolver(
            simple_no_adm_version).resolve()
        self.assertEqual(len(result), 1)

        #  * with translations
        with_trans_version = get_version_for_object(
            self.with_translation)
        result = RecursiveRevisionConflictResolver(with_trans_version).resolve()
        # one version for the object itself, and 2 for translations
        self.assertEqual(len(result), 3)

        #  * with placeholders
        with_ph_version = get_version_for_object(self.with_placeholder)
        result = RecursiveRevisionConflictResolver(with_ph_version).resolve()
        self.assertEqual(len(result), 1)

        #  * with fks
        simple_fk_version = get_version_for_object(self.simple_fk)
        result = RecursiveRevisionConflictResolver(simple_fk_version).resolve()
        self.assertEqual(len(result), 1)

        #  * with all of that
        complex_fk_version = get_version_for_object(self.complex_one_fk)
        result = RecursiveRevisionConflictResolver(complex_fk_version).resolve()
        # one version for the object itself, and 2 for translations
        self.assertEqual(len(result), 3)

        # test with conflicts
        #  * with simple object
        self.simple_no_admin.delete()
        self.assertEqual(SimpleNoAdmin.objects.count(), 0)
        self.assertEqual(SimpleFK.objects.count(), 0)
        result = RecursiveRevisionConflictResolver(simple_fk_version).resolve()
        # one for simple_fk one for simple_fk's relation to simple_no_admin
        self.assertEqual(len(result), 2)

        #  * with translations,
        #  * with placeholders,
        #  * with FKs
        with_ph_ph_pk = self.with_placeholder.content.pk
        complex_fk_ph_pk = self.complex_one_fk.complex_content.pk
        self.with_placeholder.content.delete()
        self.complex_one_fk.complex_content.delete()
        self.assertEqual(Placeholder.objects.filter(
            pk__in=[with_ph_ph_pk, complex_fk_ph_pk]).count(), 0)

        self.with_placeholder.delete()
        self.assertEqual(WithPlaceholder.objects.count(), 0)
        self.assertEqual(ComplexOneFK.objects.count(), 0)
        result = RecursiveRevisionConflictResolver(complex_fk_version).resolve()
        # 2 for translations, 2 for placeholders, 2 for deleted objects.
        self.assertEqual(len(result), 6)

        # test with version and prepared to resolve.
        # for second and third if clauses.
        self.assertEqual(MultiLevelFK.objects.count(), 0)
        multi_level_fk_version = get_version_for_object(
            self.multi_level_fk)
        result = RecursiveRevisionConflictResolver(
            multi_level_fk_version, [simple_fk_version]).resolve()
        # expecting:
        # [<Version: MultiLevelFK object>, <Version: ComplexOneFK object>,
        # <Version: Complex Content>, <Version: English>, <Version: German>,
        # <Version: WithPlaceholder object>, <Version: Helper model Content>,
        # <Version: SimpleFK object>, <Version: SimpleNoAdmin object>]
        self.assertEqual(len(result), 9)
        # test with duplicates in to_resolve. use results from
        # resolving_conflicts for one of the dependencies.
        result = RecursiveRevisionConflictResolver(
            multi_level_fk_version,
            RecursiveRevisionConflictResolver(simple_fk_version).resolve()
        ).resolve()
        self.assertEqual(len(result), 9)



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from django.core.exceptions import PermissionDenied
from django.core.urlresolvers import reverse, NoReverseMatch
from django.contrib import messages
try:
    from django.contrib.admin.utils import unquote
except ImportError:
    # Django<=1.6
    import warnings
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", DeprecationWarning)
        from django.contrib.admin.util import unquote
from django.contrib.admin.templatetags.admin_urls import add_preserved_filters
from django.db import transaction
from django.http import HttpResponseRedirect
from django.shortcuts import get_object_or_404, render_to_response
from django.template import RequestContext
from django.utils.encoding import force_text
from django.utils.translation import ugettext as _

from cms.admin.placeholderadmin import PlaceholderAdminMixin
import reversion
from reversion.models import Version
from reversion.admin import VersionAdmin

from .core import create_revision
from .forms import RecoverObjectWithTranslationForm
from .utils import (
    get_conflict_fks_versions, build_obj_repr,
    get_deleted_placeholders_for_object, object_is_translation,
    get_translation_info_message, RecursiveRevisionConflictResolver,
)


class VersionedPlaceholderAdminMixin(PlaceholderAdminMixin, VersionAdmin):
    revision_confirmation_template = 'aldryn_reversion/confirm_reversion.html'
    recover_confirmation_template = 'aldryn_reversion/confirm_recover.html'

    def add_plugin(self, request):
        with transaction.atomic():
            return super(VersionedPlaceholderAdminMixin, self).add_plugin(
                request)

    def edit_plugin(self, request, plugin_id):
        with transaction.atomic():
            return super(VersionedPlaceholderAdminMixin, self).edit_plugin(
                request, plugin_id)

    def move_plugin(self, request):
        with transaction.atomic():
            return super(VersionedPlaceholderAdminMixin, self).move_plugin(
                request)

    def delete_plugin(self, request, plugin_id):
        with transaction.atomic():
            return super(
                VersionedPlaceholderAdminMixin, self).delete_plugin(
                request, plugin_id)

    def get_commen_plugin_info(self, plugin):
        """
        Returns a dict with plugin info (to use in comment for revision)
        """
        return {'plugin_id': plugin.id, 'plugin': force_text(plugin)}

    def _create_aldryn_revision(self,
                                plugin, user=None,
                                comment=None, source=None):
        #
        # _get_attached_objects returns the models which define the
        # PlaceholderField to which this placeholder is linked. Theoretically it
        # is possible to have a placeholder attached to multiple models (as two
        # PlaceholderFields could point to the same instance), but the only way
        # to do this is by coding it. As we don't support this use case yet,
        # better to fail loudly than to compromise the integrity of the data by
        # applying the versioning to the wrong model.
        #

        get_attached_object = self._get_placeholder_attached_object

        obj_from_target = get_attached_object(plugin.placeholder)

        if source:
            obj_from_source = get_attached_object(source)
        else:
            obj_from_source = None

        if not obj_from_target and not obj_from_source:
            return

        if obj_from_target:
            create_revision(obj_from_target, user=user, comment=comment)

        if obj_from_source and obj_from_source != obj_from_target:
            create_revision(obj_from_source, user=user, comment=comment)

    def _get_placeholder_attached_object(self, placeholder):
        objs = placeholder._get_attached_objects()

        # _get_attached_objects returns the models which define the
        # PlaceholderField to which this placeholder is linked. Theoretically it
        # is possible to have a placeholder attached to multiple models (as two
        # PlaceholderFields could point to the same instance), but the only way
        # to do this is by coding it. As we don't support this use case yet,
        # better to fail loudly than to compromise the integrity of the data by
        # applying the versioning to the wrong model.

        assert len(objs) <= 1, 'Placeholder attached to multiple objects'

        try:
            obj = objs[0]
        except IndexError:
            obj = None
        return obj

    def post_add_plugin(self, request, placeholder, plugin):
        super(VersionedPlaceholderAdminMixin, self).post_add_plugin(
            request, placeholder, plugin)
        comment_dict = self.get_commen_plugin_info(plugin)
        comment = _('Added plugin #%(plugin_id)s: %(plugin)s') % comment_dict
        self._create_aldryn_revision(plugin, request.user, comment)

    def post_edit_plugin(self, request, plugin):
        super(VersionedPlaceholderAdminMixin, self).post_edit_plugin(
            request, plugin)
        comment_dict = self.get_commen_plugin_info(plugin)
        comment = _('Edited plugin #%(plugin_id)s: %(plugin)s') % comment_dict
        self._create_aldryn_revision(plugin, request.user, comment)

    def post_move_plugin(self, request, source_placeholder, target_placeholder,
                         plugin):
        super(VersionedPlaceholderAdminMixin, self).post_move_plugin(
            request, source_placeholder, target_placeholder, plugin)
        comment_dict = self.get_commen_plugin_info(plugin)
        comment = _('Moved plugin #%(plugin_id)s: %(plugin)s') % comment_dict
        self._create_aldryn_revision(
            plugin,
            request.user,
            comment,
            source=source_placeholder
        )

    def post_delete_plugin(self, request, plugin):
        super(VersionedPlaceholderAdminMixin, self).post_delete_plugin(
            request, plugin)
        comment_dict = self.get_commen_plugin_info(plugin)
        comment = _('Deleted plugin #%(plugin_id)s: %(plugin)s') % comment_dict
        self._create_aldryn_revision(plugin, request.user, comment)

    def log_addition(self, request, obj):
        """
        Override reversion.VersionAdmin log addition to provide useful message.
        """
        super(reversion.VersionAdmin, self).log_addition(request, obj)
        comment = _(
            "Initial version of %(object_repr)s.%(translation_info)s") % {
                'object_repr': build_obj_repr(obj),
                'translation_info': get_translation_info_message(obj)}
        # previous implementation was to use self.get_revision_data
        # but that was also removed in 1.9.0 since it was a duplicate of logic
        # that is already present in save_revision or its related calls.
        self.revision_manager.save_revision(
            [obj],
            user=request.user,
            comment=comment,
            ignore_duplicates=self.ignore_duplicate_revisions,
            db=self.revision_context_manager.get_db(),
        )

    def log_change(self, request, obj, message, deletion=False):
        # prepare correct change message so that we can distinguish which
        # revision would be restored. if object has language code - apppend
        # it to the message, but if previous operation was translation deletion
        # do not modify the message, it is already prepared.
        if not deletion:
            message = "{0} {1}{2}".format(
                message, build_obj_repr(obj), get_translation_info_message(obj))
        super(VersionedPlaceholderAdminMixin, self).log_change(
            request, obj, message)

    # TODO: extract to separate translation admin mixin
    def log_deletion(self, request, obj, object_repr):
        # skip not translation objects, we don't need to do anything on regular
        # objects
        if not object_is_translation(obj):
            super(VersionedPlaceholderAdminMixin, self).log_deletion(
                request, obj, object_repr)
            return

        # django-reversion does not provides you with log deletion,
        # for recover view it just uses diff between real objects
        # and the last revision objects for given model.
        # Instead wev will use log_change for translation master object
        message = _(
            "Translation deletion for %(object_repr)s "
            "('%(lang_code)s' language).") % {
                'object_repr': build_obj_repr(obj.master),
                'lang_code': obj.language_code.upper()}
        self.log_change(request, obj.master, message, deletion=True)

    @transaction.atomic
    def revision_view(self, request, object_id, version_id,
                      extra_context=None):
        if not self.has_change_permission(request):
            raise PermissionDenied()

        obj = get_object_or_404(self.model, pk=unquote(object_id))
        version = get_object_or_404(Version, pk=unquote(version_id),
                                    object_id=force_text(obj.pk))
        revision = version.revision

        if request.method == "POST":
            revision.revert()
            opts = self.model._meta
            pk_value = obj._get_pk_val()
            preserved_filters = self.get_preserved_filters(request)

            msg_dict = {
                'name': force_text(opts.verbose_name),
                'obj': force_text(obj)
            }
            msg = _('The %(name)s "%(obj)s" was successfully reverted. '
                    'You may edit it again below.') % msg_dict
            self.message_user(request, msg, messages.SUCCESS)
            redirect_url = reverse(
                'admin:%s_%s_change' % (opts.app_label, opts.model_name),
                args=(pk_value,),
                current_app=self.admin_site.name
            )
            redirect_url = add_preserved_filters({
                'preserved_filters': preserved_filters,
                'opts': opts,
            }, redirect_url)
            return HttpResponseRedirect(redirect_url)
        else:
            context = {
                'object': obj,
                'version': version,
                'revision': revision,
                'revision_date': revision.date_created,
                'versions': revision.version_set.order_by(
                    'content_type__model', 'object_id_int').all,
                'object_name': force_text(self.model._meta.verbose_name),
                'app_label': self.model._meta.app_label,
                'opts': self.model._meta,
                'add': False,
                'change': True,
                'save_as': False,
                'has_add_permission': self.has_add_permission(request),
                'has_change_permission': self.has_change_permission(
                    request, obj),
                'has_delete_permission': self.has_delete_permission(
                    request, obj),
                'has_file_field': True,
                'has_absolute_url': False,
                'original': obj,
            }
            return render_to_response(self.revision_confirmation_template,
                                      context, RequestContext(request))

    @transaction.atomic
    def recover_view(self, request, version_id, extra_context=None):
        if not self.has_change_permission(request):
            raise PermissionDenied()

        version = get_object_or_404(Version, pk=unquote(version_id))
        obj = version.object_version.object
        revision = version.revision

        # check for conflicts, it is better that user would solve them
        conflict_fks_versions = get_conflict_fks_versions(
            obj, version, revision)

        # build urls to point user onto restore links for conflicts
        opts = self.model._meta
        non_reversible_by_user = []
        conflicts_links_to_restore = []
        for fk_version in conflict_fks_versions:
            # try to point user to conflict recover views
            try:
                link = reverse(
                    'admin:{0}_{1}_recover'.format(
                        opts.app_label,
                        fk_version.object_version.object._meta.model_name),
                    args=[fk_version.pk])
                link_dict = {
                    'version': fk_version,
                    'link': link
                }
            except NoReverseMatch:
                # if there is exception either model is not registered
                # with VersionedPlaceholderAdminMixin or there is no admin
                # for that model. In both cases we need to revert this object
                # to avoid conflicts / integrity errors
                non_reversible_by_user.append(fk_version)
            else:
                conflicts_links_to_restore.append(link_dict)

        # check if we need to restore placeholder fields
        object_placeholders = get_deleted_placeholders_for_object(obj, revision)
        # if there are conflicts that cannot be resolved manually by the user
        # rely on resolver.
        if len(non_reversible_by_user) > 0:
            non_reversible_by_user = RecursiveRevisionConflictResolver(
                non_reversible_by_user[0], non_reversible_by_user[1:]).resolve()

        # prepare form kwargs
        restore_form_kwargs = {
            'revision': revision,
            'obj': obj,
            'version': version,
            'resolve_conflicts': non_reversible_by_user,
            'placeholders': object_placeholders
        }

        if request.method == "POST":
            form = RecoverObjectWithTranslationForm(request.POST,
                                                    **restore_form_kwargs)
            # form.is_valid would perform validation against foreign keys
            if form.is_valid():
                # form save will restore desired versions for object and its
                # translations
                form.save()
                # FIXME: optimize response
                opts = self.model._meta
                pk_value = obj._get_pk_val()
                preserved_filters = self.get_preserved_filters(request)

                msg_dict = {
                    'name': force_text(opts.verbose_name),
                    'obj': force_text(obj)
                }
                msg = _('The %(name)s "%(obj)s" was successfully recovered. '
                        'You may edit it again below.') % msg_dict
                self.message_user(request, msg, messages.SUCCESS)
                redirect_url = reverse(
                    'admin:%s_%s_change' % (opts.app_label, opts.model_name),
                    args=(pk_value,),
                    current_app=self.admin_site.name
                )
                # TODO: Check if there is next parameter and redirect to
                # next, for cases of conflict solving.
                redirect_url = add_preserved_filters({
                    'preserved_filters': preserved_filters,
                    'opts': opts,
                }, redirect_url)
                return HttpResponseRedirect(redirect_url)
        else:
            # populate form with regular data
            form = RecoverObjectWithTranslationForm(**restore_form_kwargs)

        context = {
            'object': obj,
            'version': version,
            'revision': revision,
            'revision_date': revision.date_created,
            'conflict_links': conflicts_links_to_restore,
            'non_resolvable_conflicts': non_reversible_by_user,
            'placeholders_to_restore': object_placeholders,
            'versions': revision.version_set.order_by(
                'content_type__name', 'object_id_int').all,
            'object_name': force_text(self.model._meta.verbose_name),
            'app_label': self.model._meta.app_label,
            'opts': self.model._meta,
            'add': False,
            'change': True,
            'save_as': False,
            'has_add_permission': self.has_add_permission(request),
            'has_change_permission': self.has_change_permission(
                request, obj),
            'has_delete_permission': self.has_delete_permission(
                request, obj),
            'has_file_field': True,
            'has_absolute_url': False,
            'original': obj,
        }
        # if there is no conflicts - add form to context.
        if not conflicts_links_to_restore:
            context['restore_form'] = form
        return render_to_response(self.recover_confirmation_template,
                                  context, RequestContext(request))



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

"""
Support for django-reversion on models with translatable fields and django-cms
placeholder fields.
"""
from functools import partial

from django.db.models.signals import post_save

from cms.models.pluginmodel import CMSPlugin
from reversion.revisions import (
    default_revision_manager, revision_context_manager, VersionAdapter)

# We would like this to not depend on Parler, but still support if it is
# available.
try:
    from parler import cache
except:
    pass


def _add_to_context(obj, manager=None, context=None):
    if manager is None:
        manager = default_revision_manager

    if context is None:
        context = default_revision_manager._revision_context_manager

    adapter = manager.get_adapter(obj.__class__)
    version_data = adapter.get_version_data(obj)
    context.add_to_context(manager, obj, version_data)


def create_revision(obj, user=None, comment=None):
    with revision_context_manager.create_revision():
        if user:
            revision_context_manager.set_user(user)
        if comment:
            revision_context_manager.set_comment(comment)

        _add_to_context(obj)

        if hasattr(obj._meta, 'placeholder_field_names'):
            add_placeholders_to_revision(instance=obj)


def add_placeholders_to_revision(
        instance, revision_manager=None, rev_ctx=None):
    """
    Manually add plugins to the revision.

    This function is an updated version of
    http://github.com/divio/django-cms/blob/develop/cms/utils/helpers.py#L34
    but instead of working on pages, works on models with placeholder
    fields.
    """

    add_to_context = partial(
        _add_to_context,
        manager=revision_manager,
        context=rev_ctx,
    )

    # Add the placeholder to the revision
    for name in instance._meta.placeholder_field_names:
        add_to_context(getattr(instance, name))

    # Add all plugins to the revision
    ph_ids = [getattr(instance, '{0}_id'.format(name))
              for name in instance._meta.placeholder_field_names]

    for plugin in CMSPlugin.objects.filter(placeholder_id__in=ph_ids):
        plugin_instance, _ = plugin.get_plugin_instance()

        if plugin_instance:
            add_to_context(plugin_instance)
        add_to_context(plugin)


class TranslatableVersionAdapterMixin(object):
    revision_manager = None

    def __init__(self, model):
        super(TranslatableVersionAdapterMixin, self).__init__(model)

        # If the model is translated with django-parler, register the
        # translation model to be tracked as well, by following all placeholder
        # fields, if any.
        if hasattr(model, '_parler_meta'):
            root_model = model._parler_meta.root_model
            self.revision_manager.register(root_model)

            # Also add the translations to the models to follow.
            self.follow = list(self.follow) + [model._parler_meta.root_rel_name]

            # And make sure that when we revert them, we update the translations
            # cache (this is normally done in the translation `save_base`
            # method, but it is not called when reverting changes).
            post_save.connect(self._update_cache, sender=root_model)

    def _update_cache(self, sender, instance, raw, **kwargs):
        """Update the translations cache when restoring from a revision."""
        if raw:
            # Raw is set to true (only) when restoring from fixtures or,
            # django-reversion
            cache._cache_translation(instance)


class PlaceholderVersionAdapterMixin(object):
    follow_placeholders = True

    def __init__(self, model):
        super(PlaceholderVersionAdapterMixin, self).__init__(model)

        # Add cms placeholders the to the models to follow.
        placeholders = getattr(model._meta, 'placeholder_field_names', None)

        if self.follow_placeholders and placeholders:
            self.follow = list(self.follow) + placeholders
            post_save.connect(self._add_plugins_to_revision, sender=model)

    def _add_plugins_to_revision(self, sender, instance, **kwargs):
        rev_ctx = self.revision_manager._revision_context_manager

        if rev_ctx.is_active() and not rev_ctx.is_managing_manually():
            add_placeholders_to_revision(
                instance=instance,
                revision_manager=self.revision_manager,
                rev_ctx=rev_ctx,
            )


class ContentEnabledVersionAdapter(TranslatableVersionAdapterMixin,
                                   PlaceholderVersionAdapterMixin,
                                   VersionAdapter):
    pass

version_controlled_content = partial(default_revision_manager.register,
    adapter_cls=ContentEnabledVersionAdapter,
    revision_manager=default_revision_manager)



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from django.contrib.contenttypes.models import ContentType
from django.db.models.fields.related import ForeignKey
from django.utils.encoding import force_text
from django.utils.translation import ugettext as _

import cms.models


def object_is_translation(obj):
    """
    Returns True if object is translation object.
    """
    related_name = getattr(obj, 'related_name', None)
    if related_name is None:
        return False
    if related_name != 'translations':
        return False
    return True


def build_obj_repr(obj):
    """
    Returns a unicode string of object Model name and its text representation.
    """
    return "{0}: '{1}'".format(force_text(obj._meta.model.__name__),
                               force_text(obj))


def get_translation_info_message(obj):
    """
    Prepares a simple translation info message i.e " ('EN' translation)"
    for given object, or an empty string.
    """
    language_message = ''
    if getattr(obj, 'language_code', None) is not None:
        language_message = _(" ('%(language_code)s' translation)") % {
            'language_code': obj.language_code.upper()}
    return language_message


def get_translations_versions_for_object(obj, revision, versions=None):
    """
    Returns a queryset of translation versions for given object, if versions
    provided - performs lookup on them instead of revision.version_set.all().
    """
    translatable = hasattr(obj, 'translations')
    if not translatable:
        return []

    if versions is None:
        versions = revision.version_set.all()
    # get translations versions
    translation_model = obj.translations.model
    translation_ct = ContentType.objects.get_for_model(translation_model)
    translation_version = versions.filter(content_type=translation_ct)
    return translation_version


def get_fk_models(obj, blank=None):
    """
    Returns a list of tuples (FK model, `blank`) for FK relations on a given
    object model. If blank is provided - filters by `blank` attribute value.
    :param obj: object to check FK relations
    :param blank: bool to filter relations by `blank` FK attrubute
    :return: list of tuples (FK model, fk.blank)
    """
    fk_relations = []
    for fk in obj._meta.fields:
        # process only FK and subclasses
        if not issubclass(type(fk), ForeignKey):
            continue
        if fk.blank == blank or blank is None:
            relation = {
                'fk_field': fk,
                'model': fk.rel.to,
                'content_type': ContentType.objects.get_for_model(fk.rel.to),
                'blank': fk.blank,
            }
            fk_relations.append(relation)
    return fk_relations


def get_deleted_objects_versions(versions):
    """
    Returns a list of version for deleted objects in given versions queryset.
    Performs excluding on versions queryset if exclude dictionary
    {field_name: value, ...} is provided.
    """
    deleted_versions = []
    for version in versions:
        if object_was_deleted(version):
            deleted_versions.append(version)
    return deleted_versions


def object_was_deleted(version):
    if version.object is None:
        # if there is no relation to object - it is delted
        return True
    # in case if GenericKey relation is broken, or object was takem
    # from cache - try to perform DB lookup. This will/should always hit db.
    # get object model class to access database for lookup
    check_obj_model = version.object._meta.model
    # if there is a match - object exists and we don't need to
    # restore it
    if check_obj_model.objects.filter(pk=version.object_id).count() == 0:
        return True
    return False


def get_conflict_fks_versions(obj, version, revision, exclude=None):
    """
    Lookup for deleted FKs for obj, expects version to be obj
    version from the same revision.
    If exclude provided - excludes based on that from versions to check.
    Expects exclude to be a dict of filter string, value i.e {'pk': 1}.
    Returns versions for deleted fks.
    """
    # TODO: get all conflicts, return a tuple/dict with required and not.
    fk_relations = get_fk_models(obj)
    versions_to_check = []
    for relation in fk_relations:
        found_versions = revision.version_set.exclude(
            pk=version.pk).filter(content_type=relation['content_type'])
        versions_to_check += list(found_versions.values_list('pk', flat=True))

    # convert to versions queryset instead of a list
    versions_to_check_qs = revision.version_set.filter(pk__in=versions_to_check)

    if exclude is not None:
        versions_to_check_qs = versions_to_check_qs.exclude(**exclude)

    conflict_fks_versions = get_deleted_objects_versions(
        versions_to_check_qs)
    return conflict_fks_versions


def object_has_placeholders(obj):
    """
    Returns True if given object has placeholder fields, False otherwise.
    """
    return cms.models.PlaceholderField in [type(field)
                                           for field in obj._meta.fields]


def get_placeholder_fields_names(obj):
    return [field.name for field in obj._meta.fields
            if type(field) == cms.models.PlaceholderField]


def get_deleted_placeholders(revision):
    """
    Lookup for deleted placeholders for given revision
    """
    placeholder_ct = ContentType.objects.get_for_model(cms.models.Placeholder)
    placeholder_versions = revision.version_set.filter(
        content_type=placeholder_ct)
    deleted_placeholders = get_deleted_objects_versions(placeholder_versions)
    return deleted_placeholders


def get_deleted_placeholders_for_object(obj, revision):
    """
    Return deleted placeholders for object given
    """
    if object_has_placeholders(obj):

        placeholders_versions = get_deleted_placeholders(revision)
        # add only placeholders that belong to this object,
        # accessing field_name itself refers to deleted object, but _id isn't.
        # Other approach would be to load object_repr and get data from there
        placeholders_pks = [getattr(obj, '{0}_id'.format(field_name))
                            for field_name in get_placeholder_fields_names(obj)
                            if getattr(obj,
                                       '{0}_id'.format(field_name),
                                       None) is not None]
        return [placeholder_version
                for placeholder_version in placeholders_versions
                if placeholder_version.object_id_int in placeholders_pks]
    return []


def exclude_resolved(to_exclude, objects):
    """
    Exclude items that are in to_exclude from objects, excepts both to be list
    of revision.models.Versions.
    """
    return [item for item in objects if item not in to_exclude]


class RecursiveRevisionConflictResolver(object):

    def __init__(self, version, to_resolve=None, exclude=None):
        self.resolved = []
        self.version = version
        if to_resolve is None:
            self.to_resolve = []
        else:
            self.to_resolve = to_resolve
        if exclude is not None:
            self.initial_exclude = exclude
        else:
            self.initial_exclude = []

    def _update_resolved(self, versions):
        """
        Update resolved versions.
        :param versions:
        :return:
        """
        for version in versions:
            if version not in self.resolved:
                self.resolved.append(version)

    def _update_to_resolve(self, other_conflicts):
        for item in other_conflicts:
            if item not in self.resolved and item not in self.initial_exclude:
                self.to_resolve.append(item)

    def resolve(self, version=None):

        if version is None:
            version = self.version

        obj = version.object_version.object
        revision = version.revision

        self._update_resolved([version])

        other_conflicts = get_conflict_fks_versions(obj, version, revision)

        deleted_placeholders = get_deleted_placeholders_for_object(obj,
                                                                   revision)
        if deleted_placeholders:
            self._update_resolved(deleted_placeholders)

        # check for translations
        translatable = hasattr(obj, 'translations')
        if translatable:
            translation_versions = get_translations_versions_for_object(
                obj, revision)
            self._update_resolved(translation_versions)

        # recalculate to resolve.
        self._update_to_resolve(other_conflicts)

        # recursion, we have everything we need
        if self.to_resolve:
            self.resolve(self.to_resolve.pop(0))

        # base case
        if not self.to_resolve:
            return self.resolved
